# Version of the library that will be used to upload to pypi
__version__ = "0.30.0.1"

import os.path
import json
import threading
from pyrfuniverse.utils.locker import Locker
from pyrfuniverse.utils.version import Version

try:
    import requests
except:
    pass


def check_for_updates():
    try:
        response = requests.get(f'https://pypi.org/pypi/pyrfuniverse/json')
        response.raise_for_status()
        data = response.json()
        current_version = Version(__version__)

        versions = Version.sorted([Version(i) for i in data['releases']], reverse=True)
        for i in versions:
            if i[0] == current_version[0] and i[1] == current_version[1] and i[2] == current_version[2]:
                if i[3] > current_version[3]:
                    print(f'\033[33mThere is a new patch version available: {i}, please consider upgrading!\033[0m')
                    break
    except Exception as e:
        print(e)


def read_config():
    if not os.path.exists(config_path):
        new_config = {"assets_path": "", "executable_file": ""}
        save_config(new_config)
    with Locker("config"):
        with open(config_path, "r", encoding="utf-8") as file:
            return json.load(file)


def save_config(config: dict):
    assert "assets_path" in config and "executable_file" in config
    with Locker("config"):
        with open(config_path, "w", encoding="utf-8") as file:
            json.dump(config, file, indent=True)


user_path = os.path.expanduser("~/.rfuniverse")
os.makedirs(user_path, exist_ok=True)
config_path = os.path.join(user_path, "config.json")
config = read_config()
assets_path = config["assets_path"]
executable_file = config["executable_file"]

try:
    update_thread = threading.Thread(target=check_for_updates)
    update_thread.daemon = True
    update_thread.start()
except:
    pass



from typing import Any, SupportsFloat
import pyrfuniverse.attributes as attr
from pyrfuniverse.envs.base_env import RFUniverseBaseEnv

try:
    import gymnasium as gym
except ImportError:
    print("This feature requires gymnasium, please install with `pip install gymnasium`")
    raise


class RFUniverseGymWrapper(RFUniverseBaseEnv, gym.Env):
    """
    RFUniverse base environment with Gym class.

    Args:
        executable_file: Str, the absolute path of Unity executable file. None for last used executable file; "@editor" for using Unity Editor.
        scene_file: Str, the absolute path of Unity scene JSON file. All JSON files locate at `StraemingAssets/SceneData` by default.
        assets: List, the list of pre-load assets. All assets in the list will be pre-loaded in Unity when the environment is initialized, which will save time during instanciating.
        graphics: Bool, True for showing GUI and False for headless mode.
        port: Int, the port for communication.
        proc_id: Int, the process id for the Unity environment. 0 for the first process, 1 for the second process, and so on.
        log_level: Int, the log level for Unity environment. 0 for no log, 1 for errors logs, 2 for warnings and errors, 3 for all only.
        ext_attr: List, the list of extended attributes. All extended attributes will be added to the environment.
        check_version: Bool, True for checking the version of the Unity environment and the pyrfuniverse library. False for not checking the version.
    """

    def __init__(
            self,
            executable_file: str = None,
            scene_file: str = None,
            assets: list = [],
            graphics: bool = True,
            port: int = 5004,
            proc_id=0,
            log_level=1,
            ext_attr: list[type(attr.BaseAttr)] = [],
            check_version: bool = True,
            communication_backend: str = "tcp",
    ):
        RFUniverseBaseEnv.__init__(
            self,
            executable_file=executable_file,
            scene_file=scene_file,
            assets=assets,
            graphics=graphics,
            port=port,
            proc_id=proc_id,
            log_level=log_level,
            ext_attr=ext_attr,
            check_version=check_version,
            communication_backend=communication_backend,
        )

    def env_step(self, count: int = 1, simulate: bool = True, collect: bool = True):
        """
        Send the messages of called functions to Unity and simulate for a step, then accept the data from Unity.

        Args:
            count: The number of steps for executing Unity simulation.
            simulate: Simulate Physics
            collect: Collect Data
        """
        RFUniverseBaseEnv.step(self, count)

    def step(self, action: gym.core.ActType) -> tuple[gym.core.ObsType, SupportsFloat, bool, bool, dict[str, Any]]:
        """
        Gym step.

        Args:
            action: gym action.
        """
        return gym.Env.step(self, action)

    def env_close(self):
        """
        Close the environment
        """
        RFUniverseBaseEnv.close(self)

    def close(self):
        """
        Close gym
        """
        gym.Env.close(self)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

import pyrfuniverse.grpc.RFUniverseGRPC_pb2 as RFUniverseGRPC__pb2


class GrpcServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Link = channel.unary_unary(
            '/RFUniverse.GrpcService/Link',
            request_serializer=RFUniverseGRPC__pb2.Empty.SerializeToString,
            response_deserializer=RFUniverseGRPC__pb2.Empty.FromString,
        )
        self.CSharpToPythonStream = channel.stream_unary(
            '/RFUniverse.GrpcService/CSharpToPythonStream',
            request_serializer=RFUniverseGRPC__pb2.BinaryMessage.SerializeToString,
            response_deserializer=RFUniverseGRPC__pb2.Empty.FromString,
        )
        self.PythonToCSharpStream = channel.unary_stream(
            '/RFUniverse.GrpcService/PythonToCSharpStream',
            request_serializer=RFUniverseGRPC__pb2.Empty.SerializeToString,
            response_deserializer=RFUniverseGRPC__pb2.BinaryMessage.FromString,
        )


class GrpcServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Link(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CSharpToPythonStream(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PythonToCSharpStream(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_GrpcServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        'Link': grpc.unary_unary_rpc_method_handler(
            servicer.Link,
            request_deserializer=RFUniverseGRPC__pb2.Empty.FromString,
            response_serializer=RFUniverseGRPC__pb2.Empty.SerializeToString,
        ),
        'CSharpToPythonStream': grpc.stream_unary_rpc_method_handler(
            servicer.CSharpToPythonStream,
            request_deserializer=RFUniverseGRPC__pb2.BinaryMessage.FromString,
            response_serializer=RFUniverseGRPC__pb2.Empty.SerializeToString,
        ),
        'PythonToCSharpStream': grpc.unary_stream_rpc_method_handler(
            servicer.PythonToCSharpStream,
            request_deserializer=RFUniverseGRPC__pb2.Empty.FromString,
            response_serializer=RFUniverseGRPC__pb2.BinaryMessage.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        'RFUniverse.GrpcService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class GrpcService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Link(request,
             target,
             options=(),
             channel_credentials=None,
             call_credentials=None,
             insecure=False,
             compression=None,
             wait_for_ready=None,
             timeout=None,
             metadata=None):
        return grpc.experimental.unary_unary(request, target, '/RFUniverse.GrpcService/Link',
                                             RFUniverseGRPC__pb2.Empty.SerializeToString,
                                             RFUniverseGRPC__pb2.Empty.FromString,
                                             options, channel_credentials,
                                             insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CSharpToPythonStream(request_iterator,
                             target,
                             options=(),
                             channel_credentials=None,
                             call_credentials=None,
                             insecure=False,
                             compression=None,
                             wait_for_ready=None,
                             timeout=None,
                             metadata=None):
        return grpc.experimental.stream_unary(request_iterator, target, '/RFUniverse.GrpcService/CSharpToPythonStream',
                                              RFUniverseGRPC__pb2.BinaryMessage.SerializeToString,
                                              RFUniverseGRPC__pb2.Empty.FromString,
                                              options, channel_credentials,
                                              insecure, call_credentials, compression, wait_for_ready, timeout,
                                              metadata)

    @staticmethod
    def PythonToCSharpStream(request,
                             target,
                             options=(),
                             channel_credentials=None,
                             call_credentials=None,
                             insecure=False,
                             compression=None,
                             wait_for_ready=None,
                             timeout=None,
                             metadata=None):
        return grpc.experimental.unary_stream(request, target, '/RFUniverse.GrpcService/PythonToCSharpStream',
                                              RFUniverseGRPC__pb2.Empty.SerializeToString,
                                              RFUniverseGRPC__pb2.BinaryMessage.FromString,
                                              options, channel_credentials,
                                              insecure, call_credentials, compression, wait_for_ready, timeout,
                                              metadata)



from pyrfuniverse.attributes.base_attr import BaseAttr
from pyrfuniverse.attributes.camera_attr import CameraAttr
from pyrfuniverse.attributes.gameobject_attr import GameObjectAttr
from pyrfuniverse.attributes.light_attr import LightAttr
from pyrfuniverse.attributes.collider_attr import ColliderAttr
from pyrfuniverse.attributes.controller_attr import ControllerAttr
from pyrfuniverse.attributes.rigidbody_attr import RigidbodyAttr
from pyrfuniverse.attributes.cloth_attr import ClothAttr
from pyrfuniverse.attributes.pointcloud_attr import PointCloudAttr
from pyrfuniverse.attributes.custom_attr import CustomAttr

attrs = {
"BaseAttr": BaseAttr,
"CameraAttr": CameraAttr,
"GameObjectAttr": GameObjectAttr,
"LightAttr": LightAttr,
"ColliderAttr": ColliderAttr,
"ControllerAttr": ControllerAttr,
"RigidbodyAttr": RigidbodyAttr,
"ClothAttr": ClothAttr,
"PointCloudAttr": PointCloudAttr,
"CustomAttr": CustomAttr
}



import platform
import socket
from pyrfuniverse.utils.rfuniverse_communicator_base import RFUniverseCommunicatorBase


class RFUniverseCommunicatorTCP(RFUniverseCommunicatorBase):
    def __init__(
            self,
            port: int = 5004,
            receive_data_callback=None,
            proc_type="editor",
    ):
        super().__init__(
            port=port,
            receive_data_callback=receive_data_callback,
            proc_type=proc_type)
        self.server = None
        self.client = None

    def online(self):
        print(f"Waiting for connections on port: {self.port}...")
        self.server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.server.bind(("localhost", self.port))
        self.server.listen(1)
        self.client, _ = self.server.accept()
        self.client.settimeout(None)
        self.client.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

        super().online()

    def close(self):
        if self.client is not None:
            self.client.close()
        if self.server is not None:
            self.server.close()

        super().close()

    def receive_bytes(self):
        data = bytearray()
        while len(data) < 4:
            temp_data = self.client.recv(4 - len(data))
            assert len(temp_data) != 0
            data.extend(temp_data)
        assert len(data) == 4
        length = int.from_bytes(data, byteorder="little", signed=False)
        if length == 0:
            return None
        buffer = bytearray()
        while len(buffer) < length:
            temp_data = self.client.recv(length - len(buffer))
            assert len(temp_data) != 0
            buffer.extend(temp_data)
        assert len(buffer) == length
        return buffer

    def send_bytes(self, data: bytes):
        if not self.connected:
            return
        length = len(data).to_bytes(4, byteorder="little", signed=False)
        send_len = 0
        while send_len < 4:
            send_len += self.client.send(length)
        send_len = 0
        while send_len < len(data):
            send_len += self.client.send(data)
        if platform == 'linux':
            self.client.setsockopt(socket.IPPROTO_TCP, socket.TCP_QUICKACK, 1)



import queue
import grpc
import pyrfuniverse.grpc.RFUniverseGRPC_pb2 as RFUniverseGRPC_pb2
import pyrfuniverse.grpc.RFUniverseGRPC_pb2_grpc as RFUniverseGRPC_pb2_grpc


class RFUniverseGrpcServer(RFUniverseGRPC_pb2_grpc.GrpcServiceServicer):
    def __init__(self):
        super().__init__()
        self.connected = False
        self.send_queue = queue.Queue()
        self.receive_queue = queue.Queue()

    def Link(self, request, context):
        return RFUniverseGRPC_pb2.Empty()

    def CSharpToPythonStream(self, request_iterator, context):
        for request in request_iterator:
            if not self.connected:
                return
            self.receive_queue.put(request.data)

    def PythonToCSharpStream(self, request, context):
        self.connected = True
        while self.connected:
            try:
                response = self.send_queue.get(timeout=1.)
            except queue.Empty:
                continue
            try:
                yield RFUniverseGRPC_pb2.BinaryMessage(data=response)
            except grpc.RpcError:
                self.connected = False
                break



from setuptools import setup
from setuptools import find_packages
import pyrfuniverse

VERSION = pyrfuniverse.__version__

setup(
    name="pyrfuniverse",
    version=VERSION,
    description="RFUniverse python interface",
    url="https://github.com/robotflow-initiative/pyrfuniverse",
    author="RobotFlow AI Team",
    author_email="robotflow@163.com",
    packages=find_packages(),
    install_requires=[
        "numpy>=1.14.1",
        "grpcio",
        "protobuf",
        "opencv-contrib-python",
        "requests"
    ],
    python_requires=">=3.10",
    entry_points={
        'console_scripts': [
            'pyrfuniverse=pyrfuniverse.entry_points:pyrfuniverse_entry_points',
        ]
    }
)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: RFUniverseGRPC.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14RFUniverseGRPC.proto\x12\nRFUniverse\"\x07\n\x05\x45mpty\"\x1d\n\rBinaryMessage\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\x32\xcb\x01\n\x0bGrpcService\x12,\n\x04Link\x12\x11.RFUniverse.Empty\x1a\x11.RFUniverse.Empty\x12\x46\n\x14\x43SharpToPythonStream\x12\x19.RFUniverse.BinaryMessage\x1a\x11.RFUniverse.Empty(\x01\x12\x46\n\x14PythonToCSharpStream\x12\x11.RFUniverse.Empty\x1a\x19.RFUniverse.BinaryMessage0\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'RFUniverseGRPC_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_EMPTY']._serialized_start=36
  _globals['_EMPTY']._serialized_end=43
  _globals['_BINARYMESSAGE']._serialized_start=45
  _globals['_BINARYMESSAGE']._serialized_end=74
  _globals['_GRPCSERVICE']._serialized_start=77
  _globals['_GRPCSERVICE']._serialized_end=280
# @@protoc_insertion_point(module_scope)



import time
from concurrent import futures
import grpc
from pyrfuniverse.utils.rfuniverse_communicator_base import RFUniverseCommunicatorBase
import pyrfuniverse.grpc.RFUniverseGRPC_pb2_grpc as RFUniverseGRPC_pb2_grpc
from pyrfuniverse.grpc.RFUniverseGRPC_Server import RFUniverseGrpcServer


class RFUniverseCommunicatorGRPC(RFUniverseCommunicatorBase):
    def __init__(
            self,
            port: int = 5004,
            receive_data_callback=None,
            proc_type="editor",
    ):
        super().__init__(
            port=port,
            receive_data_callback=receive_data_callback,
            proc_type=proc_type,
        )
        self.grpc_server = None
        self.rfuniverse_grpc_server = None

    def online(self):
        print(f"Waiting for connections on port: {self.port}...")
        self.grpc_server = grpc.server(futures.ThreadPoolExecutor(max_workers=5))
        self.rfuniverse_grpc_server = RFUniverseGrpcServer()
        RFUniverseGRPC_pb2_grpc.add_GrpcServiceServicer_to_server(self.rfuniverse_grpc_server,
                                                                  self.grpc_server)
        self.grpc_server.add_insecure_port(f'localhost:{self.port}')
        self.grpc_server.start()
        while not self.rfuniverse_grpc_server.connected:
            time.sleep(0.1)

        super().online()

    def close(self):
        if self.grpc_server is not None:
            self.grpc_server.stop(0)
            self.rfuniverse_grpc_server.connected = False

        super().close()

    def receive_bytes(self):
        return self.rfuniverse_grpc_server.receive_queue.get()

    def send_bytes(self, data: bytes):
        self.rfuniverse_grpc_server.send_queue.put(data)
